@@ -1,15 +1,14 @@
 use std::borrow::Cow;
 use std::error;
 use std::fmt;
+use std::future::Future;
 use std::marker::PhantomData;
 use std::ops::{Deref, DerefMut};
+use std::pin::Pin;
 use std::time::Duration;
 
-use async_trait::async_trait;
-
 use crate::inner::PoolInner;
 use crate::internals::Conn;
-pub use crate::internals::State;
 
 /// A generic connection pool.
 pub struct Pool<M>
@@ -45,11 +44,6 @@ impl<M: ManageConnection> Pool<M> {
         Builder::new()
     }
 
-    /// Returns information about the current state of the pool.
-    pub fn state(&self) -> State {
-        self.inner.state()
-    }
-
     /// Retrieves a connection from the pool.
     pub async fn get(&self) -> Result<PooledConnection<'_, M>, RunError<M::Error>> {
         self.inner.get().await
@@ -60,7 +54,11 @@ impl<M: ManageConnection> Pool<M> {
     /// Using an owning `PooledConnection` makes it easier to leak the connection pool. Therefore, [`Pool::get`]
     /// (which stores a lifetime-bound reference to the pool) should be preferred whenever possible.
     pub async fn get_owned(&self) -> Result<PooledConnection<'static, M>, RunError<M::Error>> {
-        self.inner.get_owned().await
+        Ok(PooledConnection {
+            conn: self.get().await?.take(),
+            pool: Cow::Owned(self.inner.clone()),
+            state: ConnectionState::Present,
+        })
     }
 
     /// Get a new dedicated connection that will not be managed by the pool.
@@ -72,6 +70,57 @@ impl<M: ManageConnection> Pool<M> {
     pub async fn dedicated_connection(&self) -> Result<M::Connection, M::Error> {
         self.inner.connect().await
     }
+
+    /// Returns information about the current state of the pool.
+    pub fn state(&self) -> State {
+        self.inner.state()
+    }
+
+    /// Adds a connection to the pool.
+    ///
+    /// If the connection is broken, or the pool is at capacity, the
+    /// connection is not added and instead returned to the caller in Err.
+    pub fn add(&self, conn: M::Connection) -> Result<(), AddError<M::Connection>> {
+        self.inner.try_put(conn)
+    }
+}
+
+/// Information about the state of a `Pool`.
+#[derive(Debug)]
+#[non_exhaustive]
+pub struct State {
+    /// The number of connections currently being managed by the pool.
+    pub connections: u32,
+    /// The number of idle connections.
+    pub idle_connections: u32,
+    /// Statistics about the historical usage of the pool.
+    pub statistics: Statistics,
+}
+
+/// Statistics about the historical usage of the `Pool`.
+#[derive(Debug, Default)]
+#[non_exhaustive]
+pub struct Statistics {
+    /// Total gets performed that did not have to wait for a connection.
+    pub get_direct: u64,
+    /// Total gets performed that had to wait for a connection available.
+    pub get_waited: u64,
+    /// Total gets performed that timed out while waiting for a connection.
+    pub get_timed_out: u64,
+    /// Total time accumulated waiting for a connection.
+    pub get_wait_time: Duration,
+    /// Total connections created.
+    pub connections_created: u64,
+    /// Total connections that were closed due to be in broken state.
+    pub connections_closed_broken: u64,
+    /// Total connections that were closed due to be considered invalid.
+    pub connections_closed_invalid: u64,
+    /// Total connections that were closed because they reached the max
+    /// lifetime.
+    pub connections_closed_max_lifetime: u64,
+    /// Total connections that were closed because they reached the max
+    /// idle timeout.
+    pub connections_closed_idle_timeout: u64,
 }
 
 /// A builder for a connection pool.
@@ -103,11 +152,12 @@ pub struct Builder<M: ManageConnection> {
 }
 
 /// bb8's queue strategy when getting pool resources
-#[derive(Debug, Clone, Copy)]
+#[derive(Debug, Default, Clone, Copy)]
 pub enum QueueStrategy {
     /// First in first out
     /// This strategy behaves like a queue
     /// It will evenly spread load on all existing connections, resetting their idle timeouts, maintaining the pool size
+    #[default]
     Fifo,
     /// Last in first out
     /// This behaves like a stack
@@ -115,12 +165,6 @@ pub enum QueueStrategy {
     Lifo,
 }
 
-impl Default for QueueStrategy {
-    fn default() -> Self {
-        QueueStrategy::Fifo
-    }
-}
-
 impl<M: ManageConnection> Default for Builder<M> {
     fn default() -> Self {
         Builder {
@@ -170,8 +214,8 @@ impl<M: ManageConnection> Builder<M> {
     ///
     /// Defaults to None.
     #[must_use]
-    pub fn min_idle(mut self, min_idle: Option<u32>) -> Self {
-        self.min_idle = min_idle;
+    pub fn min_idle(mut self, min_idle: impl Into<Option<u32>>) -> Self {
+        self.min_idle = min_idle.into();
         self
     }
 
@@ -199,7 +243,8 @@ impl<M: ManageConnection> Builder<M> {
     ///
     /// Will panic if `max_lifetime` is 0.
     #[must_use]
-    pub fn max_lifetime(mut self, max_lifetime: Option<Duration>) -> Self {
+    pub fn max_lifetime(mut self, max_lifetime: impl Into<Option<Duration>>) -> Self {
+        let max_lifetime = max_lifetime.into();
         assert_ne!(
             max_lifetime,
             Some(Duration::from_secs(0)),
@@ -220,7 +265,8 @@ impl<M: ManageConnection> Builder<M> {
     ///
     /// Will panic if `idle_timeout` is 0.
     #[must_use]
-    pub fn idle_timeout(mut self, idle_timeout: Option<Duration>) -> Self {
+    pub fn idle_timeout(mut self, idle_timeout: impl Into<Option<Duration>>) -> Self {
+        let idle_timeout = idle_timeout.into();
         assert_ne!(
             idle_timeout,
             Some(Duration::from_secs(0)),
@@ -335,7 +381,6 @@ impl<M: ManageConnection> Builder<M> {
 }
 
 /// A trait which provides connection-specific functionality.
-#[async_trait]
 pub trait ManageConnection: Sized + Send + Sync + 'static {
     /// The connection type this manager deals with.
     type Connection: Send + 'static;
@@ -343,15 +388,17 @@ pub trait ManageConnection: Sized + Send + Sync + 'static {
     type Error: fmt::Debug + Send + 'static;
 
     /// Attempts to create a new connection.
-    async fn connect(&self) -> Result<Self::Connection, Self::Error>;
+    fn connect(&self) -> impl Future<Output = Result<Self::Connection, Self::Error>> + Send;
     /// Determines if the connection is still connected to the database.
-    async fn is_valid(&self, conn: &mut Self::Connection) -> Result<(), Self::Error>;
+    fn is_valid(
+        &self,
+        conn: &mut Self::Connection,
+    ) -> impl Future<Output = Result<(), Self::Error>> + Send;
     /// Synchronously determine if the connection is no longer usable, if possible.
     fn has_broken(&self, conn: &mut Self::Connection) -> bool;
 }
 
 /// A trait which provides functionality to initialize a connection
-#[async_trait]
 pub trait CustomizeConnection<C: Send + 'static, E: 'static>:
     fmt::Debug + Send + Sync + 'static
 {
@@ -360,8 +407,11 @@ pub trait CustomizeConnection<C: Send + 'static, E: 'static>:
     ///
     /// The default implementation simply returns `Ok(())`. If this method returns an
     /// error, it will be forwarded to the configured error sink.
-    async fn on_acquire(&self, _connection: &mut C) -> Result<(), E> {
-        Ok(())
+    fn on_acquire<'a>(
+        &'a self,
+        _connection: &'a mut C,
+    ) -> Pin<Box<dyn Future<Output = Result<(), E>> + Send + 'a>> {
+        Box::pin(async { Ok(()) })
     }
 }
 
@@ -372,6 +422,7 @@ where
 {
     pool: Cow<'a, PoolInner<M>>,
     conn: Option<Conn<M::Connection>>,
+    pub(crate) state: ConnectionState,
 }
 
 impl<'a, M> PooledConnection<'a, M>
@@ -382,27 +433,17 @@ where
         Self {
             pool: Cow::Borrowed(pool),
             conn: Some(conn),
+            state: ConnectionState::Present,
         }
     }
 
-    pub(crate) fn drop_invalid(mut self) {
-        let _ = self.conn.take();
+    pub(crate) fn take(mut self) -> Option<Conn<M::Connection>> {
+        self.state = ConnectionState::Extracted;
+        self.conn.take()
     }
 }
 
-impl<M> PooledConnection<'static, M>
-where
-    M: ManageConnection,
-{
-    pub(crate) fn new_owned(pool: PoolInner<M>, conn: Conn<M::Connection>) -> Self {
-        Self {
-            pool: Cow::Owned(pool),
-            conn: Some(conn),
-        }
-    }
-}
-
-impl<'a, M> Deref for PooledConnection<'a, M>
+impl<M> Deref for PooledConnection<'_, M>
 where
     M: ManageConnection,
 {
@@ -413,7 +454,7 @@ where
     }
 }
 
-impl<'a, M> DerefMut for PooledConnection<'a, M>
+impl<M> DerefMut for PooledConnection<'_, M>
 where
     M: ManageConnection,
 {
@@ -422,7 +463,7 @@ where
     }
 }
 
-impl<'a, M> fmt::Debug for PooledConnection<'a, M>
+impl<M> fmt::Debug for PooledConnection<'_, M>
 where
     M: ManageConnection,
     M::Connection: fmt::Debug,
@@ -432,15 +473,29 @@ where
     }
 }
 
-impl<'a, M> Drop for PooledConnection<'a, M>
+impl<M> Drop for PooledConnection<'_, M>
 where
     M: ManageConnection,
 {
     fn drop(&mut self) {
-        self.pool.as_ref().put_back(self.conn.take());
+        if let ConnectionState::Extracted = self.state {
+            return;
+        }
+
+        debug_assert!(self.conn.is_some(), "incorrect state {:?}", self.state);
+        if let Some(conn) = self.conn.take() {
+            self.pool.as_ref().put_back(conn, self.state);
+        }
     }
 }
 
+#[derive(Debug, Clone, Copy)]
+pub(crate) enum ConnectionState {
+    Present,
+    Extracted,
+    Invalid,
+}
+
 /// bb8's error type.
 #[derive(Debug, Clone, PartialEq, Eq)]
 pub enum RunError<E> {
@@ -483,6 +538,33 @@ where
     }
 }
 
+/// Error type returned by `Pool::add(conn)`
+#[derive(Debug, Clone, PartialEq, Eq)]
+pub enum AddError<C> {
+    /// The connection was broken before it could be added.
+    Broken(C),
+    /// Unable to add the connection to the pool due to insufficient capacity.
+    NoCapacity(C),
+}
+
+impl<E: error::Error + 'static> fmt::Display for AddError<E> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        match *self {
+            AddError::Broken(_) => write!(f, "The connection was broken before it could be added"),
+            AddError::NoCapacity(_) => write!(
+                f,
+                "Unable to add the connection to the pool due to insufficient capacity"
+            ),
+        }
+    }
+}
+
+impl<E: error::Error + 'static> error::Error for AddError<E> {
+    fn source(&self) -> Option<&(dyn error::Error + 'static)> {
+        None
+    }
+}
+
 /// A trait to receive errors generated by connection management that aren't
 /// tied to any particular caller.
 pub trait ErrorSink<E>: fmt::Debug + Send + Sync + 'static {
